# queue-infusemedia - a Workers Framework

Usage:

Queue can be used as a dependency library or as an independent node.

## Independent node usage

1. Install: `git clone --branch 1.0.0 git@github.com:insegment/queue-infusemedia.git`
2. Configure
3. Declare queues: `php entry.php declare`
3. Run: `php entry.php manage`

## Library usage

Install:

1. Add to your project's composer.json

```
"repositories": [{
   "type": "vcs",
   "url": "git@github.com:insegment/queue-infusemedia.git"
}],
"require": {
   "insegment/queue-infusemedia": "^1.0.0"
}
```

2. Anywhere in your project you create an entry point to the queue-infusemedia (below we assume workers.php), like that:

```
require_once __DIR__ . '/vendor/autoload.php';

$appDir = __DIR__;
$configFileName = 'config.yaml';
$app = new \InSegment\WorkersFramework\Console\Application($appDir, $configFileName);
$app->handle();
```

You can specify any $appDir and any $configFileName you want, you may run whatever code you want anywhere before the initialization of the queue-infusemedia application

3. Configure
4. Declare queues: `php workers.php declare`
3. Run: `php workers.php manage`

## Just publish message example
(driver-dependent so far)

```
require_once __DIR__ . '/vendor/autoload.php';

$just = new \InSegment\WorkersFramework\Drivers\RabbitMq\JustPublishMessage(__DIR__, 'config.yaml');
$just->publishMessage('mq', 'queue_m0', 'message', ['job_name' => 'test_job']);
```

## Configuration

### Http server

```
http_port: 443
http_user: user
http_password: password
http_https: true
http_https_cert: /path/to/cert.pem
```

These options are for HTTP/HTTPS server for managing the queue. Quite nothing to say.

### Run command

```
worker_run_command: php entry.php work %s
```

This is run command by default, when using as a library, this should be a command to start worker with name %s, so to start the worker, it should go through an Application entry. If your entry is workers.php, worker_run_command should be changed accordingly

### Connections

In the connections section there must be preconfigured a connection to the queue server. Every connection must list all the queues which are mentioned in the config file and belong to the connected server. Queue names there are provided only for example, they doesn't need to match specific naming conditions.

```
connections:
  mq:
    driver: rabbit
    host: rabbit-host.example.com
    vhost: rabbit-vhost
    user: rabbit-user
    password: rabbit-password
    queues:
      - queue_m0
      - queue_m0_service
      - w0_service
      - w0_work
      - w0_dlx
```

It is possible to have queries on multiple connections, maybe workers on a local in-memory queue, and the input source of jobs is a different remote cloud instance.

Example:


```
connections:
  mq_remote:
    driver: rabbit
    host: rabbit-host.example.com
    vhost: rabbit-vhost
    user: rabbit-user
    password: rabbit-password
    queues:
      - queue_m0
      - queue_m0_service
  mq_local:
    driver: rabbit
    host: localhost
    vhost: rabbit-localvhost
    user: rabbit-localuser
    password: rabbit-localpassword
    queues:
      - w0_service
      - w0_work
      - w0_dlx
```

Work and dlx queue of each worker must be on the same connection

### Node options

These options configure how the node of the queue-infusemedia behaves

```
main_queue: queue_m0
service_queue: queue_m0_service
```

From the main_queue the Workers Framework reads the jobs and dispatches the job to the workers. A service queue is used to listen the state changes, which can be produced by nodes, or manually, for example, to turn off the queue.

The node itself is identified by its name
```
node_name: m0
```

This name is used for workers to set on which nodes they may run. Then a job set to run on these workers would run on the nodes on which run these workers.

The state file stores the status of workers - whether they are locked, whether the queue is locked
```
state_file: storage/state.yaml
```

### Multi-master

Workers Framework could be run in the way when multiple managers distribute jobs over the same set of workers, or over a different set of workers managed centrally, or in a mixed combination of shared and separate workers.

In multi-master configuration, the manager and an execution of the jobs are distributed, and the status of the work being done is united. The distributed manager can accept jobs from multiple sources less than or equal to a number of nodes in the cluster. Each node has its own source configured as main_queue, and its own service_queue.

Each node in multi-master configuration must know about each other node's service_queue to be able to share status. If the nodes are clustered in such manner, any node can turn on/off the consuming of all the nodes or some of workers, so it is possible to build a central point of control for the cluster.

To know about each other, nodes use the entry in the configuration:

```
multi_master:
  m1:
    service_queue: queue_m1_service
```

If you don't need multi-master configuration, just omit the whole entry.

Every node is the cluster must know all the workers of all the nodes in the cluster, and their configs about workers must match. Configuration of jobs similarly to workers must be known to all the nodes, unless some nodes consume a different source, which wouldn't have the some jobs, and its workers are dedicated. Then these workers are only known to the other nodes to be able to control them centrally, and the jobs may not be known to them, as the other nodes would never see these jobs

### Workers

Each worker is a process, which consumes jobs. Several options could be set there:

```
workers:
  w0:
    service_queue: w0_service
    work_queue: w0_work
    dlx_queue: w0_dlx
    dlx_requeue: true
    dlx_ttl: 5000
  W1:
    service_queue: W1_service
    work_queue: W1_work
    nodes: [m1]
```

**work_queue** of the worker is a queue from where the worker reads the jobs, which were distributed to it
**service_queue** of the worker is a queue to communicate between the worker and its manager

In multi-master configuration, the same worker may have different **work_queue** depending on which node it runs, or may have the same one. It is not really necessary for the same workers on the different nodes to have different **service_queue**, as the N nodes would normally produce N shutdown signals which will be consumed by workers when the worker is shut down. However, it may be configured the way that they have different **service_queue** on each node, too.

**dlx_queue** is a queue where the messages which were not processed properly are gone. It may be configured that the worker sends such messages to some other queue and if the **dlx_requeue** is also set it will return to the original work_queue of the worker after **dlx_ttl**

**nodes** is an array which defines on which nodes a worker can run. If the option is omitted, a worker can run on all the nodes. For workers which cannot be run on the node, no process would be started.

### Jobs

Each job must be an implementor of the interface **\InSegment\WorkersFramework\Interfaces\JobInterface**

```
jobs:
  test_job: 
    class: Workers\ExampleJob
    workers: [w0]
  test_job2:
    class: Workers\ExampleJob
  test_job3:
    class: Workers\ExampleJob
    workers: [W1]
```

**class** of the job defines the class which must be visible in the loaded namespaces of the project. If run as a library, you can usually choose any class visible in your project or its autoloaded dependencies. If run as a standalone node, you may like to add some additional namespaces to the composer.json autoload section.

A job can be limited to be run on some **workers**. If a worker is also limited to be run on a certain node, it would effectively mean, that the job would always be run on a certain node. It is possible to have groups of nodes, where jobs and workers does not interfere.
