<?php

namespace InSegment\WorkersFramework\Drivers\RabbitMq;

use Bunny\Channel;

use InSegment\WorkersFramework\Interfaces\QueueMessageInterface;

abstract class ChannelDriver
{
    
    /**
     * RabbitMq channel
     *
     * @var \Bunny\Channel
     */
    protected $channel;
    
    /**
     * Constructor
     * 
     * @param \Bunny\Channel $channel
     */
    public function __construct(Channel $channel)
    {
        $this->channel = $channel;
    }
    
    /**
     * Publish message to the queue
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueMessageInterface $message
     * @return bool|int|\React\Promise\PromiseInterface
     */
    public function publish(QueueMessageInterface $message)
    {
        return $this->channel->publish(
            $message->getContents(), // String
            $message->getHeaders()->all(), // Headers
            '', // Exchange name
            $message->getQueueName() // Routing key
        );
    }
    
    /**
     * Un-subscribe from channel
     * 
     * @param string $consumerTag
     * @return bool|\Bunny\Protocol\MethodBasicCancelOkFrame|\React\Promise\PromiseInterface<\Bunny\Protocol\MethodBasicCancelOkFrame>
     */
    public function unsubscribe(string $consumerTag)
    {
        return $this->channel->cancel($consumerTag);
    }
}
